/*
 * <license>
 * 
 * uBench v1.0
 * 
 * This software is provided to enhance knowledge and encourage progress in the scientific
 * community. It should be used only for research and educational purposes. Any reproduction
 * or use for commercial purpose, public redistribution, in source or binary forms, with or 
 * without modifications, is NOT ALLOWED without the previous authorization of the copyright 
 * holder. The origin of this software must not be misrepresented; you must not claim that you
 * wrote the original software. If you use this software for any purpose (e.g. publication),
 * a reference to the software package and the authors must be included.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Copyright (c) 2012-2014, Trasgo Group, Universidad de Valladolid.
 * All rights reserved.
 * 
 * More information on http://trasgo.infor.uva.es/
 * 
 * @author Yuri Torres (yuri.torres@infor.uva.es)
 * @author Arturo Gonzalez-Escribano (arturo@infor.uva.es)
 * 
 * </license>
 */

/* 
 * uBench v1.0_tc
 *
 * Modified in December 2015:
 * - Refactored all kernels into a single CUDA file
 * - Rewrote kernels to use one dimentional grids and blocks
 * - Recreated missing benchmarks 13 & 14 from literature
 *
 * @author Chris Quackenbush (chrisq@nyu.edu)
 */

// Iterations of the overload loop.
#define ITERATIONS 2000

__global__ void uBench_0(int *A, int matrix_side){
  size_t threadNum = blockIdx.x * blockDim.x + threadIdx.x;
  if (threadNum >= matrix_side * matrix_side) return;
  /* Empty kernel*/
}

__global__ void uBench_1(int *A, int matrix_side){
  size_t threadNum = blockIdx.x * blockDim.x + threadIdx.x;
  if (threadNum >= matrix_side * matrix_side) return;

  int halfBlock = blockDim.x / 2;
  int block_start_index = blockIdx.x * blockDim.x;
  size_t index = block_start_index
      + threadIdx.x % halfBlock
      + blockDim.x * (threadIdx.x / halfBlock)
      - halfBlock * (blockIdx.x % 2);

  A[index]= 99;
}

__global__ void uBench_2(int *A, int matrix_side){
  size_t threadNum = blockIdx.x * blockDim.x + threadIdx.x;
  if (threadNum >= matrix_side * matrix_side) return;

  A[threadNum]= 99;
}

__global__ void uBench_3(int *A, int matrix_side){
  size_t threadNum = blockIdx.x * blockDim.x + threadIdx.x;
  if (threadNum >= matrix_side * matrix_side) return;

  int halfBlock = blockDim.x / 2;
  int block_start_index = blockIdx.x * blockDim.x;
  size_t index = block_start_index
      + threadIdx.x % halfBlock
      + blockDim.x * (threadIdx.x / halfBlock)
      - halfBlock * (blockIdx.x % 2);

  /* Index and local variables */
  size_t i;
  int localValue = 0;
  /* Overload loop */
  for(i = 0; i < ITERATIONS; i++) {
    localValue += i % threadIdx.x;
  }
  A[index] = localValue;
}

__global__ void uBench_4(int *A, int matrix_side){
  size_t threadNum = blockIdx.x * blockDim.x + threadIdx.x;
  if (threadNum >= matrix_side * matrix_side) return;

  /* Index and local variables */
  size_t i;
  int localValue = 0;
  /* Overload loop */
  for(i = 0; i < ITERATIONS; i++) {
    localValue += i % threadIdx.x;
  }
  A[threadNum] = localValue;
}

#define VIRTUALWINDOW 2000

__global__ void uBench_5(int *A, int matrix_side){
  size_t threadNum = blockIdx.x * blockDim.x + threadIdx.x;
  if (threadNum >= matrix_side * matrix_side) return;

  int FP = threadNum % VIRTUALWINDOW + (threadNum / VIRTUALWINDOW);

  if (FP * matrix_side + FP >= matrix_side * matrix_side) return;

  /* Each block writes to a submatrix alon the diagonal. */
  A[FP * matrix_side + FP] = 99;
}

__global__ void uBench_6(int *A, int matrix_side){
  size_t threadNum = blockIdx.x * blockDim.x + threadIdx.x;
  if (threadNum >= matrix_side * matrix_side) return;

  /* Only one thread per block writes to the array */
  if (threadIdx.x == 0) {
    A[threadNum] = 99;
  }
}

__global__ void uBench_7(int *A, int matrix_side){
  size_t threadNum = blockIdx.x * blockDim.x + threadIdx.x;
  if (threadNum >= matrix_side * matrix_side) return;

  /* Variables*/
  size_t i;
  int localValue = 0;
  /* Overload loop */
  for(i = 0; i < ITERATIONS; i++) {
    localValue += i % threadIdx.x;
  }

  /* Every thread writes to the same index */
  A[0] = localValue;
}

__global__ void uBench_8(int *A, int matrix_side){
  size_t threadNum = blockIdx.x * blockDim.x + threadIdx.x;
  if (threadNum >= matrix_side * matrix_side) return;

  /* Index and local variables */
  size_t i;
  int sum_aux = 0;
  /* Local variables to access to output matrix */
  int row = threadNum / matrix_side;
  int col = threadNum % matrix_side;

  /* Sum all elements in this column. */
  for (i = 0; i < matrix_side; i++) {
    sum_aux += A[i * matrix_side + col];
  }

  A[row * matrix_side + col] = sum_aux;
}

__global__ void uBench_9(int *A, int matrix_side){
  size_t threadNum = blockIdx.x * blockDim.x + threadIdx.x;
  if (threadNum >= matrix_side * matrix_side) return;

  /* Index and local variables */
  size_t i;
  int sum_aux = 0;
  /* Local variables to access to output matrix */
  int row = threadNum / matrix_side;
  int col = threadNum % matrix_side;

  /* Sum all elements in this row. */
  for (i = 0; i < matrix_side; i++) {
    sum_aux += A[row * matrix_side + i];
  }

  A[row * matrix_side + col] = sum_aux;
}

__global__ void uBench_10(int *A, int matrix_side){
  size_t threadNum = blockIdx.x * blockDim.x + threadIdx.x;
  if (threadNum >= matrix_side * matrix_side) return;

  /* Index and local variables */
  size_t i;
  int sum_aux = 0;
  /* Local variables to access to output matrix */
  int row = threadNum / matrix_side;
  int col = threadNum % matrix_side;

  /* Sum all elements in this column. */
  for (i = 0; i < matrix_side; i++) {
    sum_aux+= A[i * matrix_side + col];
  }

  /* Overload loop */
  for(i=0; i< ITERATIONS; i++){
    sum_aux += i%threadIdx.x;
  }

  A[row * matrix_side + col] = sum_aux;
}

__global__ void uBench_11(int *A, int matrix_side){
  size_t threadNum = blockIdx.x * blockDim.x + threadIdx.x;
  if (threadNum >= matrix_side * matrix_side) return;

  /* Index and local variables */
  size_t i;
  int sum_aux = 0;
  /* Local variables to access to output matrix */
  int row = threadNum / matrix_side;
  int col = threadNum % matrix_side;

  /* Sum all elements in this row. */
  for (i = 0; i < matrix_side; i++) {
    sum_aux+= A[row * matrix_side + i];
  }

  /* Overload loop*/
  for(i=0; i< ITERATIONS; i++){
    sum_aux += i%threadIdx.x;
  }

  A[row * matrix_side + col] = sum_aux;
}

__global__ void uBench_12(int *A, int matrix_side){
  size_t threadNum = blockIdx.x * blockDim.x + threadIdx.x;
  if (threadNum >= matrix_side * matrix_side) return;
  /* Variables*/
  size_t i;
  int localValue = 0;

  /* Sum all elements in the thread block. */
  for (i = 0; i < blockDim.x; i++){
    localValue += A[blockIdx.x * blockDim.x + i];
  }

  A[threadNum]= localValue;
}

__global__ void uBench_13(int *A, int matrix_side){
  size_t threadNum = blockIdx.x * blockDim.x + threadIdx.x;
  if (threadNum >= matrix_side * matrix_side) return;

  /* Index and local variables */
  size_t i;
  int sum_aux= 0;

  /* Sum all elements in the matrix. */
  size_t matrix_length = matrix_side * matrix_side;
  size_t start_index = threadNum;
  for (i = 0; i < matrix_length; i++) {
    sum_aux+= A[(start_index + i) % matrix_length];
  }

  A[threadNum] = sum_aux;
}

__global__ void uBench_14(int *A, int matrix_side){
  size_t threadNum = blockIdx.x * blockDim.x + threadIdx.x;
  if (threadNum >= matrix_side * matrix_side) return;

  /* Index and local variables */
  size_t i;
  int sum_aux= 0;

  /* Sum all elements in the matrix. */
  size_t matrix_length = matrix_side * matrix_side;
  size_t start_index = threadNum * 32;
  for (i = 0; i < matrix_length; i++) {
    sum_aux+= A[(start_index + i) % matrix_length];
  }

  A[threadNum] = sum_aux;
}
