/*
 * <license>
 * 
 * uBench v1.0
 * 
 * This software is provided to enhance knowledge and encourage progress in the scientific
 * community. It should be used only for research and educational purposes. Any reproduction
 * or use for commercial purpose, public redistribution, in source or binary forms, with or 
 * without modifications, is NOT ALLOWED without the previous authorization of the copyright 
 * holder. The origin of this software must not be misrepresented; you must not claim that you
 * wrote the original software. If you use this software for any purpose (e.g. publication),
 * a reference to the software package and the authors must be included.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Copyright (c) 2012-2014, Trasgo Group, Universidad de Valladolid.
 * All rights reserved.
 * 
 * More information on http://trasgo.infor.uva.es/
 * 
 * @author Yuri Torres (yuri.torres@infor.uva.es)
 * @author Arturo Gonzalez-Escribano (arturo@infor.uva.es)
 * 
 * </license>
 */

/* 
 * uBench v1.0_tc
 *
 * Modified in December 2015:
 * - Refactored all benchmark main functions into a single executable
 * - Launches benchmark kernels with one dimentional grids and blocks
 *
 * @author Chris Quackenbush (chrisq@nyu.edu)
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>
#include <cuda.h>
#include "kernels.cu"

const void* kernels[] = {
  (const void*)uBench_0,
  (const void*)uBench_1,
  (const void*)uBench_2,
  (const void*)uBench_3,
  (const void*)uBench_4,
  (const void*)uBench_5,
  (const void*)uBench_6,
  (const void*)uBench_7,
  (const void*)uBench_8,
  (const void*)uBench_9,
  (const void*)uBench_10,
  (const void*)uBench_11,
  (const void*)uBench_12,
  (const void*)uBench_13,
  (const void*)uBench_14
};

// Usage: ./ubench <benchmark number> <thread block size> <matrix dimention>
int main (int argc, char *argv[]){
  int benchmark_num = 0;
  if (argc > 1) {
    benchmark_num = atoi(argv[1]);
  }
  int block_size = 1024;
  if (argc > 2) {
    block_size = atoi(argv[2]);
  }
  int matrix_side = 96;
  if (argc > 3) {
    matrix_side = atoi(argv[3]);
  }
  int num_elements = matrix_side * matrix_side;
  int grid_size = (num_elements + block_size - 1) / block_size;

  size_t matrix_size = sizeof(int) * num_elements;

  /* Matrix creation */
  int *A = (int *)malloc(matrix_size);
  if (A == NULL)  {
    perror("Couldn't allocate matrix.\n");
    exit(EXIT_FAILURE);
  }
  int i;
  for (i = 0; i < num_elements; i++) {
    A[i] = 0;
  }

  /* Device matrix */
  int *dA;
  cudaMalloc((void**)&dA, matrix_size);

  /* Copying the matrix elements to device memory*/
  cudaMemcpy(dA, A, matrix_size, cudaMemcpyHostToDevice);

  /* Time variables */
  float time;       
  cudaEvent_t start, stop;
  cudaEventCreate(&start);
  cudaEventCreate(&stop);
  cudaEventRecord(start, 0);

  /* Kernel launch*/
  void* args[] = { &dA, &matrix_side };
  cudaFuncSetCacheConfig(kernels[benchmark_num], cudaFuncCachePreferL1);
  cudaLaunchKernel(kernels[benchmark_num], dim3(grid_size), dim3(block_size), args, 0, 0);

  /* Calculating the time*/
  cudaEventRecord(stop, 0);
  cudaEventSynchronize(stop);
  cudaEventElapsedTime(&time, start, stop);

  /* Copying the matrix elements host memory */  
  cudaMemcpy(A, dA, matrix_size, cudaMemcpyDeviceToHost);

  /* Printing the execution kernel time*/
  printf("**** El tiempo del kernel es %f\n",time);

  /* Free the different memories and events*/
  free(A);
  cudaFree(dA);
  cudaEventDestroy(start);
  cudaEventDestroy(stop);

  /* Ending the CUDA threads */  
  cudaThreadExit();
  
  return EXIT_SUCCESS;
}








